#global imports
from db.entities.layer_adapter import LayerAdapter
from services.layer_adapter_service import LayerAdapterService

import json
from flask import Response, request

def all():
    return LayerAdapterService.all()

def add(name:str):
    '''
    add an empty layer to the DB

    @params:
        name - Required : unique identifier for the layer
    '''
    if LayerAdapterService.one(name) != None:
        return Response(status=400, response=f"Layer with name '{name}' already exists!")
    
    LayerAdapterService.add(name)
    return Response(status=200)

def add_complete():
    '''
    add a layer already containing attribute mappings to the BD.
    '''
    data = request.json

    if "name" not in data or "properties" not in data or "cluster_properties" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (name, properties, cluster_properties)")
    
    layer = LayerAdapterService.one(data["name"])

    if layer != None:
        return Response(status=400, response=f'Layer with name "{data["name"]}" already exists!')
    
    try:
        layer_new = LayerAdapter.from_serializable_dict(data)
        LayerAdapterService.add_complete(layer_new)
        return Response(status=200)
    except:
        return Response(status=400)
    
def one(name:str):
    '''
    fetch a single layer from the DB

    @params:
        name - Required : unique identifier for the layer
    '''
    layer = LayerAdapterService.one(name)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")
    
    return Response(status=200, response=json.dumps(layer.to_serializable_dict()))

def add_mapping(name:str):
    '''
    add a new mapping to the layer identified by name

    @params:
        name - Required : unique identifier for the layer
    '''
    layer = LayerAdapterService.one(name)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")
    
    data = request.json

    if "internal" not in data or "external" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (internal, external)")
    
    layer.add_mapping(data["internal"], data["external"])
    LayerAdapterService.update(layer)

    return Response(status=200)

def delete_mapping(name:str):
    '''
    delete a mapping from the layer identified by name

    @params:
        name - Required : unique identifier for the layer
    '''
    layer = LayerAdapterService.one(name)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")
    
    data = request.json

    if "internal" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (internal)")
    
    try:
        layer.delete_mapping(data["internal"])
        LayerAdapterService.update(layer)
    except ValueError:
        return Response(status=400, response=f'{data["internal"]} is not a property of the layer!')

    return Response(status=200)

def add_cluster_mapping(name:str):
    '''
    add a mapped property to the list of properties to cluster with

    @params:
        name - Required : unique identifier for the layer
    '''
    layer = LayerAdapterService.one(name)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")
    
    data = request.json

    if "attribute" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (attribute)")
    
    try:
        layer.add_cluster_mapping(data["attribute"])
        LayerAdapterService.update(layer)
        return Response(status=200)
    except:
        return Response(status=400, response=f'{data["attribute"]} is no attribute in the layer!')

def delete_cluster_mapping(name:str):
    '''
    remove a mapped property from the list of properties to cluster with

    @params:
        name - Required : unique identifier for the layer
    '''
    layer = LayerAdapterService.one(name)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")
    
    data = request.json

    if "attribute" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (attribute)")
    
    try:
        layer.delete_cluster_mapping(data["attribute"])
        LayerAdapterService.update(layer)
        return Response(status=200)
    except ValueError as e:
        print(e)
        return Response(status=400, response=f'{data["attribute"]} is no attribute in the layer!')

def delete_one(name:str):
    '''
    delete a layer and all its mappings from the Db

    @params:
        name - Required : unique identifier for the layer
    '''
    layer = LayerAdapterService.one(name)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")
    
    LayerAdapterService.delete(layer)
    return Response(status=200)